import os
import shutil
import socket
import tempfile
from concurrent import futures
from contextlib import closing
from datetime import datetime
from typing import List, Tuple
from urllib.parse import urlparse

import grpc
import numpy as np
import pandas as pd
import pytest
from google.protobuf.duration_pb2 import Duration
from pandas.util.testing import assert_frame_equal
from pyspark.sql import DataFrame, SparkSession
from pyspark.sql.types import (
    BooleanType,
    DoubleType,
    IntegerType,
    StructField,
    StructType,
    TimestampType,
)
from pytz import utc

from feast import Client, Entity, Feature, FeatureTable, FileSource, ValueType
from feast.core import CoreService_pb2_grpc as Core
from feast.data_format import ParquetFormat
from feast.pyspark.abc import SparkJobStatus
from tests.feast_core_server import CoreServicer


def find_free_port():
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        s.bind(("", 0))
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return s.getsockname()[1]


free_port = find_free_port()


def assert_dataframe_equal(left: DataFrame, right: DataFrame):
    is_column_equal = set(left.columns) == set(right.columns)

    if not is_column_equal:
        print(f"Column not equal. Left: {left.columns}, Right: {right.columns}")
    assert is_column_equal

    is_content_equal = (
        left.exceptAll(right).count() == 0 and right.exceptAll(left).count() == 0
    )
    if not is_content_equal:
        print("Rows are different.")
        print("Left:")
        left.show()
        print("Right:")
        right.show()

    assert is_content_equal


@pytest.yield_fixture(scope="module")
def spark():
    spark_session = (
        SparkSession.builder.appName("Historical Feature Retrieval Test")
        .master("local")
        .getOrCreate()
    )
    yield spark_session
    spark_session.stop()


@pytest.fixture()
def server():
    server = grpc.server(futures.ThreadPoolExecutor(max_workers=10))
    Core.add_CoreServiceServicer_to_server(CoreServicer(), server)
    server.add_insecure_port(f"[::]:{free_port}")
    server.start()
    yield server
    server.stop(0)


@pytest.fixture()
def client(server):
    return Client(core_url=f"localhost:{free_port}")


@pytest.yield_fixture()
def client_with_local_spark(tmpdir):
    import pyspark

    spark_staging_location = f"file://{os.path.join(tmpdir, 'staging')}"
    historical_feature_output_location = (
        f"file://{os.path.join(tmpdir, 'historical_feature_retrieval_output')}"
    )

    return Client(
        core_url=f"localhost:{free_port}",
        spark_launcher="standalone",
        spark_standalone_master="local",
        spark_home=os.path.dirname(pyspark.__file__),
        spark_staging_location=spark_staging_location,
        historical_feature_output_location=historical_feature_output_location,
        historical_feature_output_format="parquet",
    )


@pytest.fixture()
def client_with_tfrecord_output(tmpdir):
    import pyspark

    spark_staging_location = f"file://{os.path.join(tmpdir, 'staging')}"
    historical_feature_output_location = (
        f"file://{os.path.join(tmpdir, 'historical_feature_retrieval_tfrecord_output')}"
    )

    return Client(
        core_url=f"localhost:{free_port}",
        spark_launcher="standalone",
        spark_standalone_master="local",
        spark_home=os.path.dirname(pyspark.__file__),
        spark_staging_location=spark_staging_location,
        historical_feature_output_location=historical_feature_output_location,
        historical_feature_output_format="tfrecord",
    )


@pytest.fixture()
def driver_entity(client):
    return client.apply(Entity("driver_id", "description", ValueType.INT32))


@pytest.fixture()
def customer_entity(client):
    return client.apply(Entity("customer_id", "description", ValueType.INT32))


def create_temp_parquet_file(
    spark: SparkSession, filename, schema: StructType, data: List[Tuple]
) -> Tuple[str, str]:
    temp_dir = tempfile.mkdtemp()
    file_path = os.path.join(temp_dir, filename)
    df = spark.createDataFrame(spark.sparkContext.parallelize(data), schema)
    df.write.parquet(file_path)
    return temp_dir, f"file://{file_path}"


@pytest.fixture()
def transactions_feature_table(spark, client):
    schema = StructType(
        [
            StructField("customer_id", IntegerType()),
            StructField("event_timestamp", TimestampType()),
            StructField("created_timestamp", TimestampType()),
            StructField("total_transactions", DoubleType()),
            StructField("is_vip", BooleanType()),
        ]
    )
    df_data = [
        (
            1001,
            datetime(year=2020, month=9, day=1, tzinfo=utc),
            datetime(year=2020, month=9, day=1, tzinfo=utc),
            50.0,
            True,
        ),
        (
            1001,
            datetime(year=2020, month=9, day=1, tzinfo=utc),
            datetime(year=2020, month=9, day=2, tzinfo=utc),
            100.0,
            True,
        ),
        (
            2001,
            datetime(year=2020, month=9, day=1, tzinfo=utc),
            datetime(year=2020, month=9, day=1, tzinfo=utc),
            400.0,
            False,
        ),
        (
            1001,
            datetime(year=2020, month=9, day=2, tzinfo=utc),
            datetime(year=2020, month=9, day=1, tzinfo=utc),
            200.0,
            False,
        ),
        (
            1001,
            datetime(year=2020, month=9, day=4, tzinfo=utc),
            datetime(year=2020, month=9, day=1, tzinfo=utc),
            300.0,
            False,
        ),
    ]
    temp_dir, file_uri = create_temp_parquet_file(
        spark, "transactions", schema, df_data
    )
    file_source = FileSource(
        event_timestamp_column="event_timestamp",
        created_timestamp_column="created_timestamp",
        file_format=ParquetFormat(),
        file_url=file_uri,
    )
    features = [
        Feature("total_transactions", ValueType.DOUBLE),
        Feature("is_vip", ValueType.BOOL),
    ]
    feature_table = FeatureTable(
        "transactions", ["customer_id"], features, batch_source=file_source
    )
    yield client.apply(feature_table)
    shutil.rmtree(temp_dir)


@pytest.fixture()
def bookings_feature_table(spark, client):
    schema = StructType(
        [
            StructField("driver_id", IntegerType()),
            StructField("event_timestamp", TimestampType()),
            StructField("created_timestamp", TimestampType()),
            StructField("total_completed_bookings", IntegerType()),
        ]
    )
    df_data = [
        (
            8001,
            datetime(year=2020, month=9, day=1, tzinfo=utc),
            datetime(year=2020, month=9, day=1, tzinfo=utc),
            100,
        ),
        (
            8001,
            datetime(year=2020, month=9, day=2, tzinfo=utc),
            datetime(year=2020, month=9, day=2, tzinfo=utc),
            150,
        ),
        (
            8002,
            datetime(year=2020, month=9, day=2, tzinfo=utc),
            datetime(year=2020, month=9, day=2, tzinfo=utc),
            200,
        ),
    ]
    temp_dir, file_uri = create_temp_parquet_file(spark, "bookings", schema, df_data)

    file_source = FileSource(
        event_timestamp_column="event_timestamp",
        created_timestamp_column="created_timestamp",
        file_format=ParquetFormat(),
        file_url=file_uri,
    )
    features = [Feature("total_completed_bookings", ValueType.INT32)]
    max_age = Duration()
    max_age.FromSeconds(86400)
    feature_table = FeatureTable(
        "bookings", ["driver_id"], features, batch_source=file_source, max_age=max_age
    )
    yield client.apply(feature_table)
    shutil.rmtree(temp_dir)


@pytest.fixture()
def bookings_feature_table_with_mapping(spark, client):
    schema = StructType(
        [
            StructField("id", IntegerType()),
            StructField("datetime", TimestampType()),
            StructField("created_datetime", TimestampType()),
            StructField("total_completed_bookings", IntegerType()),
        ]
    )
    df_data = [
        (
            8001,
            datetime(year=2020, month=9, day=1, tzinfo=utc),
            datetime(year=2020, month=9, day=1, tzinfo=utc),
            100,
        ),
        (
            8001,
            datetime(year=2020, month=9, day=2, tzinfo=utc),
            datetime(year=2020, month=9, day=2, tzinfo=utc),
            150,
        ),
        (
            8002,
            datetime(year=2020, month=9, day=2, tzinfo=utc),
            datetime(year=2020, month=9, day=2, tzinfo=utc),
            200,
        ),
    ]
    temp_dir, file_uri = create_temp_parquet_file(spark, "bookings", schema, df_data)

    file_source = FileSource(
        event_timestamp_column="datetime",
        created_timestamp_column="created_datetime",
        file_format=ParquetFormat(),
        file_url=file_uri,
        field_mapping={"id": "driver_id"},
    )
    features = [Feature("total_completed_bookings", ValueType.INT32)]
    max_age = Duration()
    max_age.FromSeconds(86400)
    feature_table = FeatureTable(
        "bookings", ["driver_id"], features, batch_source=file_source, max_age=max_age
    )
    yield client.apply(feature_table)
    shutil.rmtree(temp_dir)


def test_historical_feature_retrieval_from_local_spark_session(
    spark,
    client,
    driver_entity,
    customer_entity,
    bookings_feature_table,
    transactions_feature_table,
):
    schema = StructType(
        [
            StructField("customer_id", IntegerType()),
            StructField("driver_id", IntegerType()),
            StructField("event_timestamp", TimestampType()),
        ]
    )
    df_data = [
        (1001, 8001, datetime(year=2020, month=9, day=1, tzinfo=utc)),
        (2001, 8001, datetime(year=2020, month=9, day=2, tzinfo=utc)),
        (2001, 8002, datetime(year=2020, month=9, day=1, tzinfo=utc)),
        (1001, 8001, datetime(year=2020, month=9, day=2, tzinfo=utc)),
        (1001, 8001, datetime(year=2020, month=9, day=3, tzinfo=utc)),
        (1001, 8001, datetime(year=2020, month=9, day=4, tzinfo=utc)),
    ]
    temp_dir, file_uri = create_temp_parquet_file(
        spark, "customer_driver_pair", schema, df_data
    )
    customer_driver_pairs_source = FileSource(
        event_timestamp_column="event_timestamp",
        file_format=ParquetFormat(),
        file_url=file_uri,
    )
    joined_df = client.get_historical_features_df(
        ["transactions:total_transactions", "bookings:total_completed_bookings"],
        customer_driver_pairs_source,
    )
    expected_joined_df_schema = StructType(
        [
            StructField("customer_id", IntegerType()),
            StructField("driver_id", IntegerType()),
            StructField("event_timestamp", TimestampType()),
            StructField("transactions__total_transactions", DoubleType()),
            StructField("bookings__total_completed_bookings", IntegerType()),
        ]
    )
    expected_joined_df_data = [
        (1001, 8001, datetime(year=2020, month=9, day=1, tzinfo=utc), 100.0, 100),
        (2001, 8001, datetime(year=2020, month=9, day=2, tzinfo=utc), 400.0, 150),
        (2001, 8002, datetime(year=2020, month=9, day=1, tzinfo=utc), 400.0, None),
        (1001, 8001, datetime(year=2020, month=9, day=2, tzinfo=utc), 200.0, 150),
        (1001, 8001, datetime(year=2020, month=9, day=3, tzinfo=utc), 200.0, 150),
        (1001, 8001, datetime(year=2020, month=9, day=4, tzinfo=utc), 300.0, None),
    ]
    expected_joined_df = spark.createDataFrame(
        spark.sparkContext.parallelize(expected_joined_df_data),
        expected_joined_df_schema,
    )
    assert_dataframe_equal(joined_df, expected_joined_df)
    shutil.rmtree(temp_dir)


def test_historical_feature_retrieval_with_field_mappings_from_local_spark_session(
    spark, client, driver_entity, bookings_feature_table_with_mapping,
):
    schema = StructType(
        [
            StructField("driver_id", IntegerType()),
            StructField("event_timestamp", TimestampType()),
        ]
    )
    df_data = [
        (8001, datetime(year=2020, month=9, day=1, tzinfo=utc)),
        (8001, datetime(year=2020, month=9, day=2, tzinfo=utc)),
        (8002, datetime(year=2020, month=9, day=1, tzinfo=utc)),
    ]
    temp_dir, file_uri = create_temp_parquet_file(spark, "drivers", schema, df_data)
    entity_source = FileSource(
        event_timestamp_column="event_timestamp",
        file_format=ParquetFormat(),
        file_url=file_uri,
    )
    joined_df = client.get_historical_features_df(
        ["bookings:total_completed_bookings"], entity_source,
    )
    expected_joined_df_schema = StructType(
        [
            StructField("driver_id", IntegerType()),
            StructField("event_timestamp", TimestampType()),
            StructField("bookings__total_completed_bookings", IntegerType()),
        ]
    )
    expected_joined_df_data = [
        (8001, datetime(year=2020, month=9, day=1, tzinfo=utc), 100),
        (8001, datetime(year=2020, month=9, day=2, tzinfo=utc), 150),
        (8002, datetime(year=2020, month=9, day=1, tzinfo=utc), None),
    ]
    expected_joined_df = spark.createDataFrame(
        spark.sparkContext.parallelize(expected_joined_df_data),
        expected_joined_df_schema,
    )
    assert_dataframe_equal(joined_df, expected_joined_df)
    shutil.rmtree(temp_dir)


@pytest.mark.usefixtures(
    "driver_entity",
    "customer_entity",
    "bookings_feature_table",
    "transactions_feature_table",
)
def test_historical_feature_retrieval_with_pandas_dataframe_input(
    client_with_local_spark,
):

    customer_driver_pairs_pandas_df = pd.DataFrame(
        np.array(
            [
                [1001, 8001, datetime(year=2020, month=9, day=1, tzinfo=utc)],
                [2001, 8001, datetime(year=2020, month=9, day=2, tzinfo=utc)],
                [2001, 8002, datetime(year=2020, month=9, day=1, tzinfo=utc)],
                [1001, 8001, datetime(year=2020, month=9, day=2, tzinfo=utc)],
                [1001, 8001, datetime(year=2020, month=9, day=3, tzinfo=utc)],
                [1001, 8001, datetime(year=2020, month=9, day=4, tzinfo=utc)],
            ]
        ),
        columns=["customer_id", "driver_id", "event_timestamp"],
    )
    customer_driver_pairs_pandas_df = customer_driver_pairs_pandas_df.astype(
        {"customer_id": "int32", "driver_id": "int32"}
    )

    job_output = client_with_local_spark.get_historical_features(
        ["transactions:total_transactions", "bookings:total_completed_bookings"],
        customer_driver_pairs_pandas_df,
    )

    output_dir = job_output.get_output_file_uri()
    joined_df = pd.read_parquet(urlparse(output_dir).path)

    expected_joined_df = pd.DataFrame(
        np.array(
            [
                [1001, 8001, datetime(year=2020, month=9, day=1), 100.0, 100],
                [2001, 8001, datetime(year=2020, month=9, day=2), 400.0, 150],
                [2001, 8002, datetime(year=2020, month=9, day=1), 400.0, None],
                [1001, 8001, datetime(year=2020, month=9, day=2), 200.0, 150],
                [1001, 8001, datetime(year=2020, month=9, day=3), 200.0, 150],
                [1001, 8001, datetime(year=2020, month=9, day=4), 300.0, None],
            ]
        ),
        columns=[
            "customer_id",
            "driver_id",
            "event_timestamp",
            "transactions__total_transactions",
            "bookings__total_completed_bookings",
        ],
    )
    expected_joined_df = expected_joined_df.astype(
        {
            "customer_id": "int32",
            "driver_id": "int32",
            "transactions__total_transactions": "float64",
            "bookings__total_completed_bookings": "float64",
        }
    )

    assert_frame_equal(
        joined_df.sort_values(
            by=["customer_id", "driver_id", "event_timestamp"]
        ).reset_index(drop=True),
        expected_joined_df.sort_values(
            by=["customer_id", "driver_id", "event_timestamp"]
        ).reset_index(drop=True),
    )


@pytest.mark.usefixtures(
    "driver_entity",
    "customer_entity",
    "bookings_feature_table",
    "transactions_feature_table",
)
def test_historical_feature_retrieval_with_tfrecord_output(
    client_with_tfrecord_output,
):

    customer_driver_pairs_pandas_df = pd.DataFrame(
        np.array(
            [
                [1001, 8001, datetime(year=2020, month=9, day=1, tzinfo=utc)],
                [2001, 8001, datetime(year=2020, month=9, day=2, tzinfo=utc)],
                [2001, 8002, datetime(year=2020, month=9, day=1, tzinfo=utc)],
                [1001, 8001, datetime(year=2020, month=9, day=2, tzinfo=utc)],
                [1001, 8001, datetime(year=2020, month=9, day=3, tzinfo=utc)],
                [1001, 8001, datetime(year=2020, month=9, day=4, tzinfo=utc)],
            ]
        ),
        columns=["customer_id", "driver_id", "event_timestamp"],
    )
    customer_driver_pairs_pandas_df = customer_driver_pairs_pandas_df.astype(
        {"customer_id": "int32", "driver_id": "int32"}
    )

    job_output = client_with_tfrecord_output.get_historical_features(
        ["transactions:total_transactions", "bookings:total_completed_bookings"],
        customer_driver_pairs_pandas_df,
    )

    job_output.get_output_file_uri()
    assert job_output.get_status() == SparkJobStatus.COMPLETED



import logging
import time
from concurrent import futures
from typing import Dict

import grpc

from feast.core import FeatureTable_pb2 as FeatureTableProto
from feast.core.CoreService_pb2 import ListFeatureTablesResponse
from feast.core.CoreService_pb2_grpc import CoreServiceStub
from feast.serving import ServingService_pb2_grpc as Serving
from feast.serving.ServingService_pb2 import GetFeastServingInfoResponse

_ONE_DAY_IN_SECONDS = 60 * 60 * 24


class ServingServicer(Serving.ServingServiceServicer):
    def __init__(self, core_url: str = None):
        if core_url:
            self.__core_channel = None
            self.__connect_core(core_url)
            self._feature_tables = (
                dict()
            )  # type: Dict[str, FeatureTableProto.FeatureTable]

    def __connect_core(self, core_url: str):
        if not core_url:
            raise ValueError("Please set Feast Core URL.")

        if self.__core_channel is None:
            self.__core_channel = grpc.insecure_channel(core_url)

        try:
            grpc.channel_ready_future(self.__core_channel).result(timeout=5)
        except grpc.FutureTimeoutError:
            raise ConnectionError(
                "connection timed out while attempting to connect to Feast Core gRPC server "
                + core_url
            )
        else:
            self._core_service_stub = CoreServiceStub(self.__core_channel)

    def __get_feature_tables_from_core(self):
        # Get updated list of feature tables
        feature_tables = (
            self._core_service_stub.ListFeatureTables
        )  # type: ListFeatureTablesResponse

        # Store each feature table locally
        for feature_table in list(feature_tables.tables):
            self._feature_tables[feature_table.name] = feature_table

    def GetFeastServingVersion(self, request, context):
        return GetFeastServingInfoResponse(version="0.10.0")


def serve():
    server = grpc.server(futures.ThreadPoolExecutor(max_workers=10))
    Serving.add_ServingServiceServicer_to_server(ServingServicer(), server)
    server.add_insecure_port("[::]:50052")
    server.start()
    try:
        while True:
            time.sleep(_ONE_DAY_IN_SECONDS)
    except KeyboardInterrupt:
        server.stop(0)


if __name__ == "__main__":
    logging.basicConfig()
    serve()



from collections import defaultdict
from concurrent import futures
from contextlib import contextmanager
from datetime import datetime

import grpc

from feast.core.JobService_pb2 import GetJobResponse
from feast.core.JobService_pb2 import Job as JobProto
from feast.core.JobService_pb2 import JobStatus, JobType
from feast.core.JobService_pb2_grpc import (
    JobServiceServicer,
    JobServiceStub,
    add_JobServiceServicer_to_server,
)
from feast.remote_job import RemoteRetrievalJob


@contextmanager
def mock_server(servicer):
    """Instantiate a helloworld server and return a stub for use in tests"""
    server = grpc.server(futures.ThreadPoolExecutor(max_workers=10))
    add_JobServiceServicer_to_server(servicer, server)
    port = server.add_insecure_port("[::]:0")
    server.start()

    try:
        with grpc.insecure_channel("localhost:%d" % port) as channel:
            yield JobServiceStub(channel)
    finally:
        server.stop(None)


class TestRemoteJob:
    def test_remote_ingestion_job(self):
        """ Test wating for the remote ingestion job to complete """

        class MockServicer(JobServiceServicer):
            """
            The RemoteJob is expected to call GetJob until its done.
            This mock JobService returns RUNNING status on the first call, and DONE on the second.
            """

            _job_statuses = [JobStatus.JOB_STATUS_DONE, JobStatus.JOB_STATUS_RUNNING]
            _call_count = defaultdict(int)

            def GetJob(self, request, context):

                self._call_count["GetJob"] += 1
                return GetJobResponse(
                    job=JobProto(
                        id="test",
                        type=JobType.RETRIEVAL_JOB,
                        status=self._job_statuses.pop(),
                        retrieval=JobProto.RetrievalJobMeta(output_location="foo"),
                    )
                )

        mock_servicer = MockServicer()
        with mock_server(mock_servicer) as service:
            remote_job = RemoteRetrievalJob(
                service, lambda: {}, "test", "foo", datetime.now(), None
            )

            assert remote_job.get_output_file_uri(timeout_sec=2) == "foo"
            assert mock_servicer._call_count["GetJob"] == 2



from pkg_resources import DistributionNotFound, get_distribution

from .client import Client
from .data_source import (
    BigQuerySource,
    FileSource,
    KafkaSource,
    KinesisSource,
    SourceType,
)
from .entity import Entity
from .feature import Feature
from .feature_table import FeatureTable
from .value_type import ValueType

try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:
    # package is not installed
    pass

__all__ = [
    "Client",
    "Entity",
    "BigQuerySource",
    "FileSource",
    "KafkaSource",
    "KinesisSource",
    "Feature",
    "FeatureTable",
    "SourceType",
    "ValueType",
]



# Copyright 2019 The Feast Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, Union

from tqdm import tqdm


class AbstractProducer:
    """
    Abstract class for Kafka producers
    """

    def __init__(self, brokers: str, row_count: int, disable_progress_bar: bool):
        self.brokers = brokers
        self.row_count = row_count

        # Progress bar will always display average rate
        self.pbar = tqdm(
            total=row_count, unit="rows", smoothing=0, disable=disable_progress_bar
        )

    def produce(self, topic: str, data: bytes):
        message = "{} should implement a produce method".format(self.__class__.__name__)
        raise NotImplementedError(message)

    def flush(self, timeout: int):
        message = "{} should implement a flush method".format(self.__class__.__name__)
        raise NotImplementedError(message)

    def _inc_pbar(self, meta):
        self.pbar.update(1)

    def _set_error(self, exception: str):
        raise Exception(exception)

    def print_results(self) -> None:
        """
        Print ingestion statistics.

        Returns:
            None: None
        """
        # Refresh and close tqdm progress bar
        self.pbar.refresh()

        self.pbar.close()

        print("Ingestion complete!")

        print(f"\nIngestion statistics:" f"\nSuccess: {self.pbar.n}/{self.row_count}")
        return None


class ConfluentProducer(AbstractProducer):
    """
    Concrete implementation of Confluent Kafka producer (confluent-kafka)
    """

    def __init__(self, brokers: str, row_count: int, disable_progress_bar: bool):
        from confluent_kafka import Producer

        self.producer = Producer({"bootstrap.servers": brokers})
        super().__init__(brokers, row_count, disable_progress_bar)

    def produce(self, topic: str, value: bytes) -> None:
        """
        Generic produce that implements confluent-kafka's produce method to
        push a byte encoded object into a Kafka topic.

        Args:
            topic (str): Kafka topic.
            value (bytes): Byte encoded object.

        Returns:
            None: None.
        """

        try:
            self.producer.produce(topic, value=value, callback=self._delivery_callback)
            # Serve delivery callback queue.
            # NOTE: Since produce() is an asynchronous API this poll() call
            #       will most likely not serve the delivery callback for the
            #       last produce()d message.
            self.producer.poll(0)
        except Exception as ex:
            self._set_error(str(ex))

        return None

    def flush(self, timeout: Optional[int]):
        """
        Generic flush that implements confluent-kafka's flush method.

        Args:
            timeout (Optional[int]): Timeout in seconds to wait for completion.

        Returns:
            int: Number of messages still in queue.
        """
        messages = self.producer.flush(timeout=timeout)
        if messages:
            raise Exception("Not all Kafka messages are successfully delivered.")
        return messages

    def _delivery_callback(self, err: str, msg) -> None:
        """
        Optional per-message delivery callback (triggered by poll() or flush())
        when a message has been successfully delivered or permanently failed
        delivery (after retries).

        Although the msg argument is not used, the current method signature is
        required as specified in the confluent-kafka documentation.

        Args:
            err (str): Error message.
            msg (): Kafka message.

        Returns:
            None
        """
        if err:
            self._set_error(err)
        else:
            self._inc_pbar(None)


class KafkaPythonProducer(AbstractProducer):
    """
    Concrete implementation of Python Kafka producer (kafka-python)
    """

    def __init__(self, brokers: str, row_count: int, disable_progress_bar: bool):
        from kafka import KafkaProducer

        self.producer = KafkaProducer(bootstrap_servers=[brokers])
        super().__init__(brokers, row_count, disable_progress_bar)

    def produce(self, topic: str, value: bytes):
        """
        Generic produce that implements kafka-python's send method to push a
        byte encoded object into a Kafka topic.

        Args:
            topic (str): Kafka topic.
            value (bytes): Byte encoded object.

        Returns:
            FutureRecordMetadata: resolves to RecordMetadata

        Raises:
            KafkaTimeoutError: if unable to fetch topic metadata, or unable
                to obtain memory buffer prior to configured max_block_ms
        """
        return (
            self.producer.send(topic, value=value)
            .add_callback(self._inc_pbar)
            .add_errback(self._set_error)
        )

    def flush(self, timeout: Optional[int]):
        """
        Generic flush that implements kafka-python's flush method.

        Args:
            timeout (Optional[int]): timeout in seconds to wait for completion.

        Returns:
            None

        Raises:
            KafkaTimeoutError: failure to flush buffered records within the
                provided timeout
        """
        messages = self.producer.flush(timeout=timeout)
        if messages:
            raise Exception("Not all Kafka messages are successfully delivered.")
        return messages


def get_producer(
    brokers: str, row_count: int, disable_progress_bar: bool
) -> Union[ConfluentProducer, KafkaPythonProducer]:
    """
    Simple context helper function that returns a AbstractProducer object when
    invoked.

    This helper function will try to import confluent-kafka as a producer first.

    This helper function will fallback to kafka-python if it fails to import
    confluent-kafka.

    Args:
        brokers (str): Kafka broker information with hostname and port.
        row_count (int): Number of rows in table

    Returns:
        Union[ConfluentProducer, KafkaPythonProducer]:
            Concrete implementation of a Kafka producer. Ig can be:
                * confluent-kafka producer
                * kafka-python producer
    """
    try:
        return ConfluentProducer(brokers, row_count, disable_progress_bar)
    except ImportError:
        print("Unable to import confluent-kafka, falling back to kafka-python")
        return KafkaPythonProducer(brokers, row_count, disable_progress_bar)



#
# Copyright 2020 The Feast Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import tempfile
from unittest.mock import patch
from urllib.parse import urlparse

import boto3
import fastavro
import pandas as pd
import pandavro
from moto import mock_s3
from pandas.testing import assert_frame_equal
from pytest import fixture

from feast.loaders.file import export_source_to_staging_location

BUCKET = "test_bucket"
FOLDER_NAME = "test_folder"
FILE_NAME = "test.avro"

LOCAL_FILE = "file:///tmp/tmp"
S3_LOCATION = f"s3://{BUCKET}/{FOLDER_NAME}"

TEST_DATA_FRAME = pd.DataFrame(
    {
        "driver": [1001, 1002, 1003],
        "transaction": [1001, 1002, 1003],
        "driver_id": [1001, 1002, 1003],
    }
)


@fixture
def avro_data_path():
    final_results = tempfile.mktemp()
    pandavro.to_avro(file_path_or_buffer=final_results, df=TEST_DATA_FRAME)
    return final_results


@patch("feast.loaders.file._get_file_name", return_value=FILE_NAME)
def test_export_source_to_staging_location_local_file_should_pass(get_file_name):
    source = export_source_to_staging_location(TEST_DATA_FRAME, LOCAL_FILE)
    assert source == [f"{LOCAL_FILE}/{FILE_NAME}"]
    assert get_file_name.call_count == 1


@mock_s3
@patch("feast.loaders.file._get_file_name", return_value=FILE_NAME)
def test_export_source_to_staging_location_dataframe_to_s3_should_pass(get_file_name):
    s3_client = boto3.client("s3")
    s3_client.create_bucket(Bucket=BUCKET)
    source = export_source_to_staging_location(TEST_DATA_FRAME, S3_LOCATION)
    file_obj = tempfile.TemporaryFile()
    uri = urlparse(source[0])
    s3_client.download_fileobj(uri.hostname, uri.path[1:], file_obj)
    file_obj.seek(0)
    avro_reader = fastavro.reader(file_obj)
    retrived_df = pd.DataFrame.from_records([r for r in avro_reader])
    assert_frame_equal(retrived_df, TEST_DATA_FRAME, check_like=True)
    assert get_file_name.call_count == 1


def test_export_source_to_staging_location_s3_file_as_source_should_pass():
    source = export_source_to_staging_location(S3_LOCATION, None)
    assert source == [S3_LOCATION]


@mock_s3
def test_export_source_to_staging_location_s3_wildcard_as_source_should_pass(
    avro_data_path,
):
    s3_client = boto3.client("s3")
    s3_client.create_bucket(Bucket=BUCKET)
    with open(avro_data_path, "rb") as data:
        s3_client.upload_fileobj(data, BUCKET, f"{FOLDER_NAME}/file1.avro")
    with open(avro_data_path, "rb") as data:
        s3_client.upload_fileobj(data, BUCKET, f"{FOLDER_NAME}/file2.avro")
    sources = export_source_to_staging_location(f"{S3_LOCATION}/*", None)
    assert sources == [f"{S3_LOCATION}/file1.avro", f"{S3_LOCATION}/file2.avro"]
